# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import namedtuple


EnvironmentVariable = namedtuple(
    'EnvironmentVariable', ('name', 'description'))
